from Player import Player
import Choice

import os,sys
import curses
import signal
import random

from urwid_reduct.curses_display import Screen

class Sprite:
    def __init__(self, chars):
	self.chars = chars
	self.height = len(chars)
	self.width = len(chars[0])

SPRITES = {
	"letters" : {
	    "Pawn" : Sprite(['P']),
	    "Knight" : Sprite(['N']),
	    "Bishop" : Sprite(['B']),
	    "Rook" : Sprite(['R']),
	    "Queen" : Sprite(['Q']),
	    "King" : Sprite(['K']),
	    },
	"3x2" : {
	    "Pawn" : Sprite([
		" o ",
		"<_>"]),

	    "Rook" : Sprite([
		"[\"]",
		"|_|"]),

	    "Knight" : Sprite([
		"/'>",
		"(_\\"]),

	    "Bishop" : Sprite([
		"(/)",
		"(_)"]),

	    "King" : Sprite([
		"\+/",
		"|_|"]),

	    "Queen" : Sprite([
		"\^/",
		"(_)"])
	    }
	}

def monoSprite(piece, player):
    char = { "Pawn" : 'p',
	    "Knight" : 'n',
	    "Bishop" : 'b',
	    "Rook" : 'r',
	    "Queen" : 'q',
	    "King" : 'k'
	    }[piece]
    if player == 0:
	return Sprite([char])
    else:
	return Sprite([char.upper()])

class CursesUI(Player):
    def __init__(self, socket, isServer, gameOpts, opts):
	Player.__init__(self, socket, isServer, gameOpts, opts)
	self.selectedSquare = (self.state.width/2,self.state.height/2)
	self.selectedPieceSquare = None
	self.highlights = []

	if opts.get("monochrome", False):
	    self.mono = True
	    self.origSqWidth = opts.get("sqWidth", 1)
	    self.origSqHeight = opts.get("sqHeight", 1)
	    self.sprites = None
	    self.sprHeight = 1
	    self.sprWidth = 1
	else:
	    self.mono = False
	    self.origSqWidth = opts.get("sqWidth", 5)
	    self.origSqHeight = opts.get("sqHeight", 3)
	    self.sprites = SPRITES[opts.get("spriteset", "letters")]
	    self.sprHeight = self.sprites["King"].height
	    self.sprWidth = self.sprites["King"].width

    def interact(self):
	signal.signal(signal.SIGWINCH, lambda n,f: self.handleSigwinch())
	signal.siginterrupt(signal.SIGWINCH, False)
	self.urwidscr = Screen()
	self.urwidscr.run_wrapper(self.wrapped_interact)

    def wrapped_interact(self):
	self.scr = self.urwidscr.s

	for bg in range(8):
	    for fg in range(8):
		try:
		    curses.init_pair(1+8*bg+fg, fg, bg)
		except:
		    pass
	curses.curs_set(2)

	self.setGeometry()

	self.blankScreen()

	self._interact()

    def setGeometry(self):
	self.height, self.width = self.scr.getmaxyx()

	self.sqHeight = self.origSqHeight
	self.sqWidth = self.origSqWidth

	self.edgeSep = 1 + (self.sqHeight <= 2)
	while self.height < self.sqHeight * (self.state.height + 2*self.edgeSep) + 2:
	    self.sqHeight -= 1
	    self.edgeSep = 1 + (self.sqHeight <= 2)
	while self.width < self.sqWidth * (self.state.width + 2*self.edgeSep) + 6:
	    self.sqWidth -= 1

	if not (self.sqHeight >= self.sprHeight and self.sqWidth >= self.sprWidth):
	    raise "Terminal too small"

	self.geomBoardStart = (self.sqWidth*self.edgeSep, self.sqHeight*self.edgeSep)
		
	self.turnCol = self.geomBoardStart[0] +\
		self.sqWidth * (self.state.width + 1) + 4
	self.scoresCol = self.turnCol + 2

	self.textRow = self.sqHeight * (self.state.height + 2*self.edgeSep) + 1

    def blankScreen(self):
	for i in range(self.width-1):
	    for j in range(self.height):
		self.scr.addch(j, i, ' ', self.colourPair(curses.COLOR_WHITE))

    def handleSigwinch(self):
	curses.endwin()
	curses.initscr()
	self.setGeometry()
	self.blankScreen()
	self.drawAll()

    def colourPair(self, fg=curses.COLOR_BLACK, bg=curses.COLOR_BLACK):
	return curses.color_pair(1+8*bg + fg)

    def turnRow(self, pl):
	if pl == 0: return self.coordsOfSquare((0, 0))[1]
	else: return self.coordsOfSquare((0, self.state.height - 1))[1]

    def rotated(self, c):
	if self.player == 1:
	    return ((self.state.height - 1) - c[0], (self.state.width - 1) - c[1])
	else:
	    return c

    # to make things slightly less ugly, we work with 'coords', which are x,y
    # pairs corresponding to the y,x pairs curses wants.
    def coordsOfSquare(self, sq):
	sq = self.rotated(sq)
	return ( self.geomBoardStart[0] +
		self.sqWidth * sq[0] +
		self.sqWidth/2,
		self.geomBoardStart[1] +
		self.sqHeight * sq[1] + self.sqHeight/2 )
    def addch(self, coord, c, attr=0):
	self.scr.addch(coord[1], coord[0], c, attr)
    def move(self, coord):
	self.scr.move(coord[1], coord[0])

    def refresh(self):
	self.move(self.coordsOfSquare(self.selectedSquare))
	self.scr.refresh()

    def drawEdges(self):
	# axes
	for i in range(self.state.width):
	    for j in [-self.edgeSep, self.state.height-1+self.edgeSep]:
		coords = list( self.coordsOfSquare((i,j)) )
		self.addch( coords,
			chr(ord('a')+i), self.colourPair(curses.COLOR_WHITE))
	for i in [-self.edgeSep, self.state.width-1+self.edgeSep]:
	    for j in range(self.state.height):
		self.addch( self.coordsOfSquare((i,j)),
			chr(ord('1')+j), self.colourPair(curses.COLOR_WHITE))


    def drawState(self):
	for i in range(self.state.width):
	    for j in range(self.state.height):
		self.drawSquare(self.state, (i, j))

    def playerColour(self, player):
	if self.mono: return curses.COLOR_WHITE
	if player == 0: return curses.COLOR_WHITE
	elif player == 1: return curses.COLOR_YELLOW
	else: return curses.COLOR_WHITE

    def drawScores(self):
	if self.scores == [0,0]: return
	for pl in [0,1]:
	    self.addch(
		(self.scoresCol, self.turnRow(pl)),
		    str(self.scores[pl]),
		    self.colourPair(curses.COLOR_WHITE))

    def gameEnd(self):
	if self.state.dead[0]:
	    if self.state.dead[1]:
		text = "Draw"
	    else:
		text = "Black wins"
	elif self.state.dead[1]:
	    text = "White wins"
	else:
	    text = "Draw"
	self.getKey(text)

    def getKey(self, prompt=""):
	self.message(prompt)
	while True:
	    ks = self.urwidscr.get_input()
	    if ks:
		key = ks[0]
		break
	self.clearMessage()
	return key

    def message(self, text=""):
	self.scr.addstr(self.textRow, 0, text)
	self.scr.clrtoeol()
	self.refresh()
    def clearMessage(self):
	self.message("")

    def drawSquare(self, state, coord):
	assert type(coord) == tuple

	boldSprite = False

	highlight = (self.selectedPieceSquare == coord or
		coord in self.highlights)
	piece = state.pieceIn(coord)

	bg = curses.COLOR_BLACK

	if piece:
	    if self.mono:
		spriteFg = curses.COLOR_WHITE
		sprite = monoSprite(piece.type, piece.player)
	    else:
		spriteFg = self.playerColour(piece.player)
		sprite = self.sprites[piece.type]
	    if highlight:
		boldSprite = True
	else:
	    spriteFg = curses.COLOR_WHITE
	    sprite = None
	spriteAttr = self.colourPair(spriteFg, bg)
	if boldSprite:
	    spriteAttr |= curses.A_BOLD

	scrCoord = self.coordsOfSquare(coord)

	for si in range(self.sqWidth):
	    for sj in range(self.sqHeight):
		charcoord = ( scrCoord[0] - self.sqWidth/2 + si,
			scrCoord[1] - self.sqHeight/2 + sj )
		if self.mono:
		    char = ' '
		    charAttr = self.colourPair(curses.COLOR_WHITE)
		else:
		    if (coord[0]+coord[1])%2 == 1:
			char = '.'
		    else:
			char = ' '
		    charAttr = self.colourPair(curses.COLOR_BLUE, bg)
		# Sprite:
		if sprite:
		    spi = si - self.sqWidth/2 + sprite.width/2
		    spj = sj - self.sqHeight/2 + sprite.height/2
		    if ( 0 <= spi < sprite.width and 
			    0 <= spj < sprite.height ):
			charAttr = spriteAttr
			char = sprite.chars[spj][spi]
		# piece highlighting (ugly)
		if False:
		    if char == ' ' and highlight:
			charAttr = spriteAttr
			if si == self.sqWidth/2 - self.sprWidth/2 - 1:
			    char = '>'
			elif si == self.sqWidth/2 - self.sprWidth/2 + self.sprWidth + 1:
			    char = '<'
		self.addch(charcoord, char, charAttr)

    def showPrevState(self):
	if self.state.lastUpdates:
	    self.state.undoUpdates(self.state.lastUpdates)
	    self.drawState()
	    self.getKey("Previous turn")
	    self.state.processStateUpdates(self.state.lastUpdates)
	    self.drawState()

    def showThreats(self):
	for (pos, piece) in self.state.board.items():
	    if self.state.threatenedBy(pos, 1-piece.player):
		self.highlights.append(pos)
	tempSel = self.selectedPieceSquare
	self.selectedPieceSquare = None
	self.drawState()
	self.getKey("Threatened pieces are highlighted")
	self.highlights = []
	self.selectedPieceSquare = tempSel
	self.drawState()

    class CancelChoiceException(BaseException):
	pass
    def getChoice(self, pl):
	self.clearMessage()
	while True:
	    try:
		primary = self.getMove(pl)
		choice = Choice.Choice(primary, None)
		while not self.state.checkLegalChoice(choice, pl):
		    self.message("Submit backup move")
		    choice = Choice.Choice(primary, self.getMove(pl))
		break
	    except self.CancelChoiceException:
		self.clearMessage()
	self.message("Waiting for opponent")
	return choice
    def getMove(self, pl):
	self.selectedPieceSquare = None
	self.drawSquare(self.state, self.selectedSquare)
	self.refresh()
	while True:
	    keys = self.urwidscr.get_input()
	    for key in keys:
		delta = [0,0]
		if key in ['6', "right", 'l', 's']:
		    delta[0] = 1
		elif key in ['4', "left", 'h', 'h']:
		    delta[0] = -1
		elif key in ['2','/','-',"down", 'j']:
		    delta[1] = 1
		elif key in ['8','*','+',"up", 'k']:
		    delta[1] = -1
		elif key in ['7', "home", 'y', 'g']:
		    delta[0] = delta[1] = -1
		elif key in ['3', "page down", 'n', 'm']:
		    delta[0] = delta[1] = 1
		elif key in ['1', "end", 'b', 'b']:
		    delta[0] = -1
		    delta[1] = 1
		elif key in ['9', "page up", 'u', 'c']:
		    delta[0] = 1
		    delta[1] = -1
		elif key == "ctrl l":
		    self.drawState()
		    self.scr.redrawwin()
		    self.scr.refresh()

		# TODO: promotion selection
		elif key in ['\n', '\r', "enter", ' ', '5']: 
		    if self.selectedPieceSquare not in [None, self.selectedSquare]:
			move = Choice.Move(self.selectedPieceSquare, self.selectedSquare)
			if self.state.checkLegalSimulMove(move, pl):
			    t = self.selectedPieceSquare
			    self.selectedPieceSquare = None
			    self.drawSquare(self.state, t)
			    return move
		    piece = self.state.pieceIn(self.selectedSquare)
		    if piece and piece.player == pl:
			if self.selectedPieceSquare == self.selectedSquare:
			    self.selectedPieceSquare = None
			    self.drawSquare(self.state, self.selectedSquare)
			else:
			    oldSelectedPiece = self.selectedPieceSquare
			    self.selectedPieceSquare = self.selectedSquare

			    if oldSelectedPiece: self.drawSquare(self.state, oldSelectedPiece)
			    self.drawSquare(self.state, self.selectedSquare)
		elif key in ['esc']:
		    t = self.selectedPieceSquare
		    self.selectedPieceSquare = None
		    if t:
			self.drawSquare(self.state, t)
		    raise self.CancelChoiceException

		elif key in ['p', 'P']:
		    self.showPrevState()

		elif key in ['t', 'T']:
		    self.showThreats()
		elif key in ['r', 'R']:
		    c = None
		    c = self.getKey("Really resign? 'Y' to confirm.")
		    if c == 'Y':
			return self.state.resignMove(pl)
		    self.clearMessage()
		    
		if delta != [0,0]:
		    if pl == 1:
			# board flipped for black
			delta[0] = -delta[0]
			delta[1] = -delta[1]
		    oldSelected = self.selectedSquare
		    self.selectedSquare = (
			    (self.selectedSquare[0] + delta[0]) % self.state.width,
			    (self.selectedSquare[1] + delta[1]) % self.state.height )
		    self.drawSquare(self.state, oldSelected)
		    self.drawSquare(self.state, self.selectedSquare)

		self.refresh()
