from Logic import VERSION

import socket
import random
import hashlib


PORT=ord('s')+(ord('c')<<7)  #12787
PROTOCOL_STRING="0000"   # must remain 4 chars

class Socket:
    """Simple interface for sockets
    Adapted from http://www.amk.ca/python/howto/sockets/sockets.html"""

    class SocketClosed(Exception): pass

    def __init__(self, sock):
	self.sock = sock
	self.recvBuffer = ""
    def send(self, msg):
	totalSent = 0
	while totalSent < len(msg):
	    sent = self.sock.send(msg[totalSent:])
	    if sent == 0:
		raise Socket.SocketClosed
	    totalSent += sent
    def recvFixed(self, msglen):
	msg = self.recvBuffer
	while len(msg) < msglen:
	    chunk = self.sock.recv(msglen-len(msg))
	    if chunk == '':
		raise Socket.SocketClosed
	    msg += chunk
	return msg
    def close(self):
	self.sock.close()


class SimulSocket(Socket):
    class BadHash(Exception): pass
    def exchangeFixedSimul(self, msg, nonceBytes=8):
	"""Send msg and "simultaneously" recieve a message of the same length,
	the illusion of simultaneity being securely achieved using secure
	hashes"""
	def makeNonce(bytes):
	    return reduce(lambda x,y: x+y,
		    [ chr(random.randrange(256)) for i in range(bytes) ],
		    "")
	nonce = makeNonce(nonceBytes)
	self.send("H: %s\n" % hashlib.sha256(msg+nonce).digest())
	assert self.recvFixed(3) == "H: "
	theirHash = self.recvFixed(32)
	assert self.recvFixed(1) == "\n"

	self.send("N: %s\n" % nonce)
	assert self.recvFixed(3) == "N: "
	theirNonce = self.recvFixed(nonceBytes)
	assert self.recvFixed(1) == "\n"

	self.send("M: %s\n" % msg)
	assert self.recvFixed(3) == "M: "
	theirMsg = self.recvFixed(len(msg))
	assert self.recvFixed(1) == "\n"

	if hashlib.sha256(theirMsg+theirNonce).digest() != theirHash:
	    # They lied to us!
	    raise SimulSocket.BadHash
	return theirMsg

    def agreedRandomBit(self):
	r1 = random.randrange(2)
	r2 = ord(self.exchangeFixedSimul(chr(r1)))
	return r1^r2

class gameSocket(SimulSocket):
    class ProtocolError(RuntimeError): pass
    def handshake(self):
	self.send(PROTOCOL_STRING)
	theirProt = self.recvFixed(len(PROTOCOL_STRING))
	if theirProt != PROTOCOL_STRING:
	    raise ProtocolError(
		    "Remote server sent bad protocol string %self" % theirProt)

	self.send(Logic.VERSION)
	theirVersion = self.recvFixed(len(Logic.VERSION))
	if theirVersion != Logic.VERSION:
	    raise ProtocolError(
		    "Remote server sent bad game logic version string %s" %
		    theirVersion)

def serve(port=PORT):
    ss = socket.socket(
	    socket.AF_INET, socket.SOCK_STREAM)
    try:
	ss.bind((socket.gethostname(), port))
	ss.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
	ss.listen(1)
	(s, address) = ss.accept()
    finally:
	ss.close()
    return gameSocket(s)

def connect(host, port=PORT):
    s = socket.socket(
	    socket.AF_INET, socket.SOCK_STREAM)
    s.connect((host, port))
    return gameSocket(s)
