from Choice import Choice, Move

TYPECODES = {
	'p' : "Pawn",
	'n' : "Knight",
	'b' : "Bishop",
	'r' : "Rook",
	'q' : "Queen",
	'k' : "King"
}

def fischerRow():
    from random import shuffle
    from re import search
    l = list("rnbkqbnr")
    while True:
	shuffle(l)
	backRow = ''.join(l)
	if search("r.*k.*r", backRow) and search("b(..)*b", backRow):
	    return backRow

class GameOpts:
    NETSTR_LEN=8
    def __init__(self, backRow = "rnbkqbnr"):
	self.backRow = backRow
    def netStr(self):
	return str(self.backRow)
    def __str__(self):
	return self.netStr()

def netStrToGameOpts(s):
    return GameOpts(s)

class State(object):
    def __init__(self, gameOpts):
	self.simultaneous = True
	self.width = 8
	self.height = 8

	self.board = {}

	backRow = gameOpts.backRow

	for pl in [0,1]:
	    for i in range(self.width):
		self.board[ (i, [self.height-1,0][pl]) ] = Piece(
			TYPECODES[backRow[i]], pl)
		self.board[ (i, [self.height-2,1][pl]) ] = Piece(
			"Pawn", pl, -(-1)**pl)

	self.dead = [False, False]
	self.enpassantable = [None, None]
	self.lastUpdates = []
	self.lastMovers = []

    def pieceIn(self, coord):
	return self.board.get(tuple(coord), None)

    def queeningRow(self, pl):
	return {0: 0, 1: 7}[pl]
    def onSide(self, c):
	return 1 - (c[1] / 4)

    def inBounds(self, c):
	return (0 <= c[0] < self.width and 0 <= c[1] < self.height)

    # All changes to state are abstracted into these StateUpdate objects, so
    # we can efficiently undo the changes (used in AI routines)
    class StateUpdate:
	def __init__(self): pass
	def undo(self): raise "BUG"
	def update(self, state): raise "BUG"
    class ChangeType(StateUpdate):
	def __init__(self, frm, oldtype, newtype):
	    self.frm, self.oldtype, self.newtype = frm, oldtype, newtype
	def undo(self):
	    return State.ChangeType(self.frm, self.newtype, self.oldtype)
	def update(self, state):
	    state.pieceIn(self.frm).type = self.newtype
    class ChangePrevPos(StateUpdate):
	def __init__(self, pos, oldprev, newprev):
	    self.pos, self.oldprev, self.newprev = pos, oldprev, newprev
	def undo(self):
	    return State.ChangePrevPos(self.pos, self.newprev, self.oldprev)
	def update(self, state):
	    #print "%s: %s -> %s" % (self.pos, self.oldprev, self.newprev)
	    state.pieceIn(self.pos).prevPos = self.newprev
    class RemovePiece(StateUpdate):
	def __init__(self, pos, piece):
	    assert piece
	    self.pos, self.piece = pos, piece.copy()
	def undo(self):
	    return State.PlacePiece(self.pos, self.piece)
	def update(self, state):
	    #print "Remove: ", self.pos, "; prevpos: ", self.piece.prevPos
	    del state.board[tuple(self.pos)]
    class PlacePiece(StateUpdate):
	def __init__(self, pos, piece):
	    self.pos, self.piece = pos, piece.copy()
	def undo(self):
	    return State.RemovePiece(self.pos, self.piece)
	def update(self, state):
	    #print "Place at ", self.pos, "; prevpos: ", self.piece.prevPos
	    assert not state.board.has_key(self.pos)
	    state.board[tuple(self.pos)] = self.piece.copy()
    class SetEnpassantability(StateUpdate):
	def __init__(self, pl, e, olde):
	    self.pl, self.e, self.olde = pl, e, olde
	def undo(self):
	    return State.SetEnpassantability(self.pl, self.olde, self.e)
	def update(self, state):
	    state.enpassantable[self.pl] = self.e
    class SetDeadState(StateUpdate):
	def __init__(self, pl, b):
	    self.pl, self.b = pl, b
	def undo(self):
	    return State.SetDeadState(self.pl, not self.b)
	def update(self, state):
	    state.dead[self.pl] = self.b

    def update(self, choices):
	moves, dominatrix = self.chooseMoves(choices)

	prevPosUpdates = []
	for (pos,piece) in self.board.items():
	    if piece.prevPos != pos:
		prevPosUpdates.append(self.ChangePrevPos(pos, piece.prevPos, pos))
	self.processStateUpdates(prevPosUpdates)
	
	self.lastMovers = []

	if all([moves[pl].frm == moves[1-pl].to for pl in [0,1]]):
	    # mutual takings
	    updates = [ self.RemovePiece(moves[pl].frm, self.pieceIn(moves[pl].frm)) for pl in [0,1] ]
	    self.processStateUpdates(updates)
	else:
	    # execute moves using normal chess rules, dominatrix first
	    # (order is important only in weirdo edge cases)
	    updates1, movedTo = self.updatesForMove(moves[dominatrix], dominatrix)
	    self.processStateUpdates(updates1)
	    updates2, _ = self.updatesForMove(moves[1-dominatrix], 1-dominatrix, movedTo)
	    self.processStateUpdates(updates2)
	    updates = updates1 + updates2

	oldUpdates = self.lastUpdates
	self.lastUpdates = prevPosUpdates + updates
	return oldUpdates

    def updatesForMove(self, move, pl, conflictSquares=[]):
	newEnpassantable = None

	mover = self.pieceIn(move.frm).copy()

	placings = []

	updates = []
	
	if move.frm == move.to and move.frm in conflictSquares:
	    # we're suiciding, but our opponent has already taken us
	    return [], []

	updates.append(self.RemovePiece(move.frm, mover))

	if mover.type == "Pawn":

	    if move.to[1] == self.queeningRow(pl):
		# queening
		mover.type = TYPECODES[move.queenPiece]

	    if move.to[0] != move.frm[0]:
		# possible en passant
		enpassantSquare = (move.to[0], move.frm[1])
		if self.enpassantable[1-pl] == enpassantSquare:
		    updates.append(self.RemovePiece(enpassantSquare, self.board[enpassantSquare]))

	    if (move.to[1] - move.frm[1]) in [2,-2]:
		# double move
		newEnpassantable = move.to

	if mover.type == "King":
	    if move.frm == move.to:
		# resignation
		updates.append(self.SetDeadState(pl, True))
		return updates, []
	    castle = self.castleAttempt(move, pl)
	    if castle:
		kingTo, rookFrm, rookTo = castle
		rook = self.pieceIn(rookFrm)
		updates.append(self.RemovePiece(rookFrm, rook))
		placings.append( (rookTo, rook) )
		move.to = kingTo

	placings.append( (move.to, mover) )
	
	movedTo = []
	for (to, piece) in placings:
	    taken = self.pieceIn(to)
	    if taken and taken.player == 1-pl:
		updates.append(self.RemovePiece(move.to, taken))
		if taken.type == "King":
		    updates.append(self.SetDeadState(1-pl, True))
	    if to in conflictSquares:
		# opponent has just moved here - both pieces get taken
		if piece.type == "King":
		    updates.append(self.SetDeadState(pl, True))
	    else:
		self.lastMovers.append(to)
		piece.moved = True
		movedTo.append(to)
		updates.append(self.PlacePiece(to, piece))

	if newEnpassantable != self.enpassantable[pl]:
	    updates.append(self.SetEnpassantability(
		pl, newEnpassantable, self.enpassantable[pl]))

	return updates, movedTo

    def chooseMoves(self, choices):
	dominatrix = 0
	moves = [choices[0].primary, choices[1].primary]
	for pl in [0,1]:
	    ours = choices[pl].primary
	    theirs = choices[1-pl].primary
	    ourmover = self.pieceIn(ours.frm)
	    theirmover = self.pieceIn(theirs.frm)
	    if ( (self.captures(theirs, 1-pl, ours.frm)
		and self.threatIsEstablished(theirs, 1-pl))
		and ours.to != theirs.frm
		and ours.to != ours.frm
		or (ourmover.type == "Pawn" and theirs.to == ours.to and
		    ours.frm[0] == ours.to[0] and
		    (theirmover.type != "Pawn" or self.onSide(ours.to) == 1-pl))):
		moves[pl] = choices[pl].secondary
		dominatrix = 1-pl
		if not moves[pl]:
		    raise BaseException("Required secondary not submitted!")

	return moves, dominatrix

    def secondaryNeeded(self, primary, pl):
	if primary.frm == primary.to:
	    # suicide
	    return False
	mover = self.pieceIn(primary.frm)
	for (pos,piece) in self.board.items():
	    if (piece.player == 1-pl
		and self.establishedThreat(Move(pos, primary.frm), 1-pl)
		and primary.to != pos):
		    return True
	if mover.type == "Pawn":
	    for (pos,piece) in self.board.items():
		if (piece.player == 1-pl
			and primary.frm[0] == primary.to[0]
			and (piece.type != "Pawn" or self.onSide(primary.to) == 1-pl)
			and self.checkLegalMove(Move(pos, primary.to), 1-pl)):
		    return True
	    if self.establishedEnpassant(primary.frm):
		return True
	return False

    def captures(self,move,pl,pos):
	# legal move by pl captures piece on square pos
	return (move.to == pos
		or (self.enpassantable[1-pl] == pos
		    and move.to == (pos[0], pos[1] - self.pieceIn(pos).direction)
		    and move.frm[0] != pos[0]
		    and self.pieceIn(move.frm).type == "Pawn"))
    def threat(self,move,pl):
	target = self.pieceIn(move.to)
	return (self.checkLegalMove(move, pl) and
		((target and target.player == 1-pl) or
		    (self.pieceIn(move.frm).type == "Pawn" and move.frm[0] != move.to[0])))
    def establishedThreat(self, move, pl):
	return self.threat(move,pl) and self.threatIsEstablished(move, pl)

    def threatIsEstablished(self, move, pl):
	prevPos = self.pieceIn(move.frm).prevPos
	if prevPos == None or self.lastUpdates == None:
	    return False
	self.undoUpdates(self.lastUpdates)
	fakeUpdates = []
	piece = self.pieceIn(move.to)
	# Put a fake enemy piece in
	if piece:
	    fakeUpdates.append(self.RemovePiece(move.to, piece))
	fakeUpdates.append(self.PlacePiece(move.to, Piece("Pawn", 1-pl, (-1)**pl)))
	self.processStateUpdates(fakeUpdates)
	ret = ( self.checkLegalMove(Move(prevPos, move.to), pl)
		or self.checkPinLine(prevPos, move.to, move.frm))
	self.undoUpdates(fakeUpdates)
	self.processStateUpdates(self.lastUpdates)
	return ret

    def threatenedBy(self, pos, pl):
	for (frm,piece) in self.board.items():
	    if (piece.player == pl
		    and self.establishedThreat(Move(frm, pos), pl)):
		return True
	if self.establishedEnpassant(pos):
	    return True
	return False

    def establishedEnpassant(self, pos):
	piece = self.pieceIn(pos)
	if not piece: return False
	pl = piece.player
	if self.enpassantable[pl] != pos: return False
	for dx in [-1,1]:
	    frm = (pos[0]+dx, pos[1])
	    if not self.inBounds(frm): continue
	    to = (pos[0], pos[1] - piece.direction)
	    oppPiece = self.pieceIn(frm)
	    if ( oppPiece and oppPiece.player == 1-pl and
		    self.establishedThreat(Move(frm,to), 1-pl) ):
		return True

    def checkLegalChoice(self, choice, pl):
	if self.secondaryNeeded(choice.primary, pl):
	    if choice.secondary == None:
		return False
	    if choice.primary.frm == choice.secondary.frm:
		return False
	    if not self.checkLegalSimulMove(choice.secondary, pl):
		return False
	return self.checkLegalSimulMove(choice.primary, pl)

    def checkLegalSimulMove(self, move, pl):
	# Check legality according to rules of simulchess
	if not self.checkLegalMove(move, pl):
	    return False
	if self.threat(move,pl) and not self.threatIsEstablished(move, pl):
	    return False
	return True
    def checkLegalMove(self, move, pl):
	# Check legality according to standard rules of chess
	mover = self.pieceIn(move.frm)
	taken = self.pieceIn(move.to)

	if mover is None or mover.player != pl:
	    return False
	if not self.inBounds(move.to):
	    return False
	if taken and taken.player == pl and not (
		mover.type=="King" and taken.type in ["Rook", "King"]):
	    return False

	def legalRun(dir):
	    checkCoord = move.frm
	    while True:
		checkCoord = (checkCoord[0] + dir[0], checkCoord[1] + dir[1])
		if checkCoord == move.to:
		    return True
		if not self.inBounds(checkCoord) or self.pieceIn(checkCoord):
		    return False

	if mover.type == "Rook":
	    for dir in [(-1,0),(1,0),(0,-1),(0,1)]:
		if legalRun(dir):
		    return True
	    return False
	if mover.type == 'Bishop':
	    for dir in [(-1,-1),(1,-1),(-1,1),(1,1)]:
		if legalRun(dir):
		    return True
	    return False
	if mover.type == 'Queen':
	    for dir in [(-1,0),(1,0),(0,-1),(0,1),(-1,-1),(1,-1),(-1,1),(1,1)]:
		if legalRun(dir):
		    return True
	    return False

	if mover.type == 'Knight':
	    for dir in [(2,-1),(1,-2),(-1,-2),(-2,-1),(-2,1),(-1,2),(1,2),(2,1)]:
		target = (move.frm[0] + dir[0], move.frm[1] + dir[1])
		if target == move.to:
		    return True
	    return False

	if mover.type == 'King':
	    if move.to == move.frm:
		# resignation
		return True
	    castle = self.castleAttempt(move, pl)
	    if castle:
		kingTo, rookFrm, rookTo = castle
		if mover.moved:
		    return False
		rook = self.pieceIn(rookFrm)
		if rook.moved:
		    return False
		interval = range(
			min(move.frm[0],kingTo[0],rookTo[0],rookFrm[0]),
			max(move.frm[0],kingTo[0],rookTo[0],rookFrm[0])+1)
		for i in interval:
		    c = (i, move.frm[1])
		    if c not in [move.frm,rookFrm] and self.pieceIn(c):
			return False
		    if self.threatenedBy(c, 1-pl):
			return False
		return True
	    elif taken and taken.player == pl:
		    return False

	    for dir in [(-1,0),(1,0),(0,-1),(0,1),(-1,-1),(1,-1),(-1,1),(1,1)]:
		target = (move.frm[0] + dir[0], move.frm[1] + dir[1])
		if target == move.to:
		    return True

	    return False

	if mover.type == "Pawn":
	    oneAhead = (move.frm[0], move.frm[1] + mover.direction)
	    twoAhead = (move.frm[0], move.frm[1] + 2*mover.direction)
	    if ( move.to == oneAhead
		    and taken is None ):
		# forward one
		return True
	    if ( move.to == twoAhead
		    and not mover.moved
		    and taken is None
		    and self.pieceIn(oneAhead) is None ):
		# forward two
		return True
	    if ( move.to[1] == oneAhead[1]
		    and move.to in [
			(oneAhead[0] + dir, oneAhead[1]) for dir in [-1,1] ]):
		# diagonal
		if taken:
		    # normal capture
		    return True
		else:
		    # check for en passant
		    enpassantSquare = (move.to[0], move.frm[1])
		    if self.enpassantable[1-pl] == enpassantSquare:
			# en passant
			return True
		    # illegal diagonal
		    return False
	    return False


	return True

    def castleAttempt(self, move, pl):
	# Check for a castle attempt, returning data if it might be one
	mover = self.pieceIn(move.frm)
	taken = self.pieceIn(move.to)
	if not mover or mover.type != "King" or move.to[1] != move.frm[1]:
	    return None
	dx = 0
	if move.to[0] == move.frm[0] + 2:
	    dx = 1
	elif move.to[0] == move.frm[0] - 2:
	    dx = -1
	elif taken and taken.player == pl and taken.type == "Rook":
		dx = (-1)**(move.to[0] < move.frm[0])
	if dx == 0:
	    return None
	if dx == 1:
	    # queenside
	    kingTo = (5, move.frm[1])
	    rookTo = (4, move.frm[1])
	else:
	    # kingside
	    kingTo = (1, move.frm[1])
	    rookTo = (2, move.frm[1])
	rook = None
	rookFrm = move.frm
	while self.inBounds(rookFrm):
	    rookFrm = (rookFrm[0]+dx, rookFrm[1])
	    rook = self.pieceIn(rookFrm)
	    if rook and rook.player == pl and rook.type == "Rook":
		break
	    else: rook=None
	if rook:
	    return kingTo, rookFrm, rookTo
	return None

    def checkPinLine(self, frm, to, inLine):
	# Look for an enemy piece which frm pins to to, such that were frm
	# moved to inLine we'd have the same pinline
	mover = self.pieceIn(frm)
	def pinRun(dir):
	    checkCoord = frm
	    pinned = None
	    lineUp = False
	    while True:
		checkCoord = (checkCoord[0] + dir[0], checkCoord[1] + dir[1])
		if checkCoord == to:
		    if pinned:
			if lineUp: return True
			checkCoord = frm
			while self.inBounds(checkCoord) and checkCoord != inLine:
			    checkCoord = (checkCoord[0] - dir[0], checkCoord[1] - dir[1])
			if checkCoord == inLine: return True
		    return False
		if not self.inBounds(checkCoord):
		    return False
		piece = self.pieceIn(checkCoord)
		if piece:
		    if not pinned and piece.player != mover.player:
			pinned = checkCoord
		    else: return False
		if checkCoord == inLine and not pinned:
		    lineUp = True
	if mover.type == "Rook":
	    for dir in [(-1,0),(1,0),(0,-1),(0,1)]:
		if pinRun(dir):
		    return True
	if mover.type == 'Bishop':
	    for dir in [(-1,-1),(1,-1),(-1,1),(1,1)]:
		if pinRun(dir):
		    return True
	if mover.type == 'Queen':
	    for dir in [(-1,0),(1,0),(0,-1),(0,1),(-1,-1),(1,-1),(-1,1),(1,1)]:
		if pinRun(dir):
		    return True
	return False


    def processStateUpdates(self, stateUpdates):
	for update in stateUpdates:
	    update.update(self)

    def undoUpdates(self, stateUpdates):
	cpy = stateUpdates[:]
	cpy.reverse()
	self.processStateUpdates(map(lambda u: u.undo(), cpy))

    def undoTurn(self, oldUpdates):
	self.undoUpdates(self.lastUpdates)
	self.lastUpdates = oldUpdates

    def possCoords(self):
	for i in range(self.width):
	    for j in range(self.height):
		yield (i,j)
    def possDests(self,frm,piece):
	def addRunDests(dir,dests):
	    dest = list(frm)
	    while True:
		dest[0] += dir[0]
		dest[1] += dir[1]
		if not self.inBounds(dest):
		    break
		dests.append(tuple(dest))

	dests = []
	if piece.type == "Pawn":
	    dy = piece.direction
	    dests = map(lambda d: (frm[0]+d[0],frm[1]+d[1]),
		    [(0,dy),(0,dy*2),(1,dy),(-1,dy)])
	elif piece.type == "King":
	    for dx in [-1,0,1]:
		for dy in [-1,0,1]:
		    dests.append((frm[0]+dx, frm[1]+dy))
	elif piece.type == "Knight":
	    for d in [(2,-1),(1,-2),(-1,-2),(-2,-1),(-2,1),(-1,2),(1,2),(2,1)]:
		dests.append((frm[0] + d[0], frm[1] + d[1]))
	elif piece.type == "Rook":
	    for dir in [(-1,0),(1,0),(0,-1),(0,1)]:
		addRunDests(dir, dests)
	elif piece.type == 'Bishop':
	    for dir in [(-1,-1),(1,-1),(-1,1),(1,1)]:
		addRunDests(dir, dests)
	elif piece.type == 'Queen':
	    for dir in [(-1,0),(1,0),(0,-1),(0,1),(-1,-1),(1,-1),(-1,1),(1,1)]:
		addRunDests(dir, dests)
	return filter(self.inBounds, dests)
    def possSimulMoves(self, pl):
	return self.possMoves(self,pl,True)
    def possMoves(self, pl, simul=False):
	for (frm,piece) in self.board.items():
	    if piece.player != pl:
		continue
	    for to in self.possDests(frm,piece):
		move = Move(frm, to)
		if ((simul and self.checkLegalSimulMove(move,pl))
			or self.checkLegalMove(move,pl)):
		    yield move
    def canMove(self, pl):
	try:
	    self.possSimulMoves().next()
	    return True
	except StopIteration:
	    return False
    def resignMove(self, pl):
	for (kingPos,king) in self.board.items():
	    if king.player == pl and king.type == "King":
		return Move(kingPos,kingPos)
	return None

class Piece(object):
    def __init__(self, type, player, direction=None, prevPos=None, moved=False):
	self.type = type
	self.player = player
	self.prevPos = prevPos
	self.moved = moved
	# for pawns - direction of movement. -1 = up; 1 = down
	self.direction = direction
    def copy(self):
	# manual copy method, since copy.copy is slow and we copy alot
	return Piece(self.type, self.player, self.direction, self.prevPos, self.moved)
